package sugar;

import javax.swing.*;
import javax.swing.border.*;
import java.util.*;
import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.text.*;
import java.awt.event.*;
import java.awt.print.*;
import javax.print.*;
import javax.swing.event.*;
import java.io.*;
import java.lang.reflect.*;
import javax.swing.event.*;

/**
 * Does not implements ActionListener
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: FEL CVUT</p>
 * @author Jiri Danecek
 * @version 1.0
 */
abstract class MyAction extends AbstractAction implements ChangeListener {

  String tip;
  KeyStroke keyStroke;
  Icon icon;

  /**
   * Empty method implementing ChangeListener
   * @param ce
   */
  public void stateChanged(ChangeEvent ce) {}

  /**
   * Constructor
   * @param msgs
   * @param icon
   * @param keyStroke
   */
  public MyAction(String [] msgs, Icon icon, KeyStroke keyStroke) {
    this(msgs[0], msgs[1], icon);
    this.keyStroke = keyStroke;
  }

  /**
   * Constructor
   * @param msgs
   * @param icon
   */
  public MyAction(String [] msgs, Icon icon) {
    this(msgs[0], msgs[1], icon);
  }

  /**
   *
   * @param title
   * @param tip
   * @param icon
   */
  public MyAction(String title, String tip, Icon icon) {
    super(title, icon);
    this.tip = tip;
    this.icon = icon;
  }

  /**
   *
   * @param titleTip
   */
  public MyAction(String[] titleTip) {
    this(titleTip[0], titleTip[1]);
  }

  /**
   *
   * @param title
   * @param tip
   */
  public MyAction(String title, String tip) {
    this(title);
    this.tip = tip;
  }

  /**
   *
   * @param title
   * @param tip
   * @param icon
   * @param keyStroke
   */
  public MyAction(String title, String tip, Icon icon, KeyStroke keyStroke) {
    this(title, icon, keyStroke);
    this.tip = tip;
  }

  /**
   *
   * @param title
   * @param keyStroke
   */
  public MyAction(String title, KeyStroke keyStroke) {
    super(title);
    this.keyStroke = keyStroke;
  }

  /**
   *
   * @param title
   * @param icon
   * @param keyStroke
   */
  public MyAction(String title, Icon icon, KeyStroke keyStroke) {
    this(title, icon);
    this.keyStroke = keyStroke;
  }

  /**
   *
   * @param title
   * @param icon
   */
  public MyAction(String title, Icon icon) { super(title, icon); }

  /**
   *
   * @param title
   */
  public MyAction(String title) { super(title); }

  /**
   *
   * @return
   */
  public String getTip() { return tip; }

  /**
   *
   * @return
   */
  KeyStroke getKeyStroke() { return keyStroke; }
}


/**
 *
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: FEL CVUT</p>
 * @author Jiri Danecek
 * @version 1.0
 */
class Caret {

  private int column;
  private Model model;
  private boolean isVisible = false;
  private ConsolePane consolePane;

  Caret(Model model, ConsolePane consolePane) {
    this.model = model;
    this.consolePane = consolePane;
  }

  private void rep() { isVisible = false; consolePane.repaint(); }

  void setPosition(int column) { this.column = column; rep(); }

  int getColumn() { return column; }

  boolean left() {
    if (column > 0) {
      column--;
      rep();
      return true;
    }
    return false;
  }

  boolean right() {
    if (column < model.inputLine.length()) {
      column++;
      rep();
      return true;
    }
    return false;
  }

  void deleteChar() {
    model.deleteCharAt(column);
    rep();
  }

  void insertChar(char c) {
    model.insertCharAt(c, column);
    column++;
    rep();
  }

  void home() { column = 0; rep(); }

  void end()  { column = model.findEnd(); rep(); }

  public void paint(Graphics g) {
    if (consolePane.hasFocus() && model.inputRequest || isVisible) {
      Graphics2D g2 = (Graphics2D)g.create();
      double lineHeight =  consolePane.fontHeight();
      double cw;
      double ch;
      Shape s;
      if (isVisible) g2.setColor(g2.getBackground());
      else g2.setColor(Color.black);
      s = new Rectangle2D.Double(
              consolePane.column2v("" + model.lastLine + model.inputLine,
              model.lastLine.length() + column),
              consolePane.row2v(model.lastRow()) - lineHeight * 0.7,
                                lineHeight * 0.05, lineHeight * 0.8);
      g2.fill(s);
      isVisible = !isVisible;
    }
  }

}


/**
 *
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: FEL CVUT</p>
 * @author Jiri Danecek
 * @version 1.0
 */
class Model {

  ArrayList content = new ArrayList();
  StringBuffer lastLine  = new StringBuffer();
  StringBuffer inputLine = new StringBuffer();
  String maxLine = "";
  ConsolePane consolePane;
  boolean inputRequest = false;

  /**
   * Sets consolePane to Model
   * @param consolePane
   */
  void setConsole(ConsolePane consolePane) { this.consolePane = consolePane; }

  /**
   *
   * @return maxLine string
   */
  String getMaxLine() { return maxLine; }

  /**
   *
   * @return
   * @throws IOException
   */
  synchronized String readLine() throws IOException {
    inputRequest = true;
    consolePane.requestFocusInWindow();
    String line;
    try { wait(); }
    catch (InterruptedException e) {
      throw new RuntimeException("Interrupted in Console.readLine()");
    }
    finally {
      line = "" + inputLine;                           // convert to string
      inputLine.setLength(0);                          // empty
      inputRequest = false;                            // end of input
      consolePane.getCaret().setPosition(0);           // caret home
    }
    return line;
  }

  /**
   * Notifies
   */
  synchronized void putEnter() { print('\n'); notifyAll(); }

  /**
   *
   * @param s
   */
  void print(String s) {for (int i=0; i<s.length(); i++) print(s.charAt(i));}

  /**
   *
   * @param c
   */
  void print(char c) {
    Thread.yield();
    if (Thread.currentThread().isInterrupted())
      throw new RuntimeException("Interrupted in Console.print(char)");
    if (c == '\n') {
      String line = "" + lastLine + inputLine;
      if (line.length() > maxLine.length())  maxLine = line;
      content.add(line);                          // add line to array
      lastLine.setLength(0);
      consolePane.revalidate();
      SwingUtilities.invokeLater(
        new Runnable() {
          public void run() {  consolePane.scrollToVisible(); }
        }
      );
    }
    else lastLine.append(c);
  }

  /**
   *
   * @param column
   */
  void deleteCharAt(int column) {
    if (inputLine.length() > column) inputLine.deleteCharAt(column);
  }

  /**
   *
   * @param c
   * @param column
   */
  void insertCharAt(char c, int column) { inputLine.insert(column, c); }

  /**
   *
   * @return
   */
  int findEnd() { return inputLine.length(); }

  /**
   *
   * @return
   */
  int rowCount() {  return lastRow() + 1; }

  /**
   *
   * @return
   */
  int lastRow() { return content.size(); }

  /**
   *
   * @param row
   * @return
   */
  String getLine(int row) {
    if (row < content.size()) return (String)content.get(row);
    else return "" + lastLine + inputLine;
  }

}

/**
 *
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: FEL CVUT</p>
 * @author Jiri Danecek
 * @version 1.0
 */

class ConsolePane extends JPanel {

  Font font = new Font("SansSerif", Font.PLAIN, 14);
  FontRenderContext frc = new FontRenderContext(new AffineTransform(), true, false);
  Model model;
  private Caret caret;

  private javax.swing.Timer timer =                            // timer
    new javax.swing.Timer(500, new ActionListener() {
      int i;
      public void actionPerformed(ActionEvent ae) { repaint(); }
    }
  );

  public int column2v(String s) { return column2v(s, s.length()); }

  public Caret getCaret() { return caret; }

  public int column2v(String s, int column) {
    if (s.length() == 0) return 0;
    return (int)font.getStringBounds(s, 0, column, frc).getWidth();
  }

  public int row2v(int row) { return (row + 1) * fontHeight(); }

  public int v2m(int v) { return  v / fontHeight(); }

  int getFontSize() { return font.getSize(); }

  int fontHeight() { return (int)font.getMaxCharBounds(frc).getHeight(); }

  Color getHighlightBackgroundColor() { return Color.lightGray; }

  Color getHighlightForegroundColor() { return Color.black; }

  public void rePaint() {
    SwingUtilities.invokeLater(new Runnable() {
      public void run() { repaint(); }
    });
  }

  public void startRefresh() { timer.start(); }

  public void stopRefresh() { timer.stop(); }

  public Dimension getPreferredSize() { return getMinimumSize(); }

  public Dimension getMinimumSize() {
    return new Dimension(column2v(model.getMaxLine()), model.rowCount() * fontHeight());
  }

  /**
   *
   * @param g
   * @param s
   * @param row
   */
  public void paintString(Graphics g, String s, int row) { //., Color color) {
    Graphics2D g2 = (Graphics2D)g.create();
    g2.setColor(this.getForeground());
    g2.drawString(s, 0, row2v(row));
  }

  /**
   *
   * @param g
   */
  public void paintComponent(Graphics g) {
    super.paintComponent(g);
    Graphics2D g2 = (Graphics2D)g;
    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    g2.setFont(font);
    for (int row = 0; row<model.rowCount(); row++)  paintString(g2, model.getLine(row), row);
    caret.paint(g2);
  }

  static Model m;

  void  scrollToVisible() { c.setPosition(); }

  Console c;

  /**
   * Constructor
   * @param model
   * @param c
   */
  public ConsolePane(final Model model, Console c) {
    this.setFocusable(true);
    this.c = c;
    this.model = model;
    caret = new Caret(model, this);

    addFocusListener(
      new FocusAdapter() {
        public void focusGained(FocusEvent e) { repaint(); }
        public void focusLost(FocusEvent e)   { repaint(); }
      }
    );

    addMouseListener(new MouseAdapter() {
      public void mousePressed(MouseEvent me) {
        if (model.inputRequest)  requestFocusInWindow();
      }
    });

    addKeyListener(
      new KeyAdapter() {
        public void keyPressed(KeyEvent ke) {
          if (!model.inputRequest) return;      // no input
          switch (ke.getKeyCode()) {            // input
            case KeyEvent.VK_DELETE: caret.deleteChar(); break;
            case KeyEvent.VK_HOME:   caret.home(); break;
            case KeyEvent.VK_END:    caret.end(); break;
            case KeyEvent.VK_ENTER: {
              model.putEnter();         // provides notification of model
              caret.home();
              scrollToVisible();
            }
          }
        }

        public void keyTyped(KeyEvent ke) {
          if (!model.inputRequest) return;
          char c = ke.getKeyChar();
          if (c == '\b') {
            if (caret.getColumn() > 0) {
              caret.left();
              caret.deleteChar();
            }
          } else {
            if (Character.isISOControl(c)) return;
            caret.insertChar(c);
          }
        }
    });
    startRefresh();
  }

  public void processKeyEvent(KeyEvent ke) {
    if (ke.getID() == KeyEvent.KEY_PRESSED) {
      switch (ke.getKeyCode()) {
        case KeyEvent.VK_LEFT:  if (caret.left()) return;
        case KeyEvent.VK_RIGHT: if (caret.right()); return;
      }
    }
    super.processKeyEvent(ke);
  }
}

/**
 *
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: FEL CVUT</p>
 * @author Jiri Danecek
 * @version 1.0
 */

public class Console extends JScrollPane {

  private static Console self;
  private ConsolePane cp;

  /**
   * Static method provides access to singleton
   * @param title
   * @return
   */
  public static Console createConsole(String title) {
    return self = new Console(title);
  }

  /**
   * Give state of Console object.
   * @return boolean.
   */
  public static boolean isInitialized() { return self != null; }

  /**
   *
   */
  void setPosition() {
    JScrollBar vsb = getVerticalScrollBar();
    vsb.setValue(vsb.getMaximum());  // !!!
  }

  /**
   * Provides access to ConsolePane.
   * @return ConsolePane
   */
  public ConsolePane getConsolePane() { return cp; }

  /**
   * Private constructor
   * @param title
   */
  private Console(String title) {
    int b = 25;
//    this.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(b,b,b,b), BorderFactory.createTitledBorder(title)));
//    this.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder(title), BorderFactory.createEmptyBorder(b,b,b,b)));
    this.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), title));
    Model model = new Model();
    cp = new ConsolePane(model, this);
    model.setConsole(cp);
    setViewportView(cp);
  }

  /**
   *
   * @param s
   */
  public static void print(String s) {
    self.cp.model.print(s);
    self.setPosition();
  }

  /**
   *
   * @param s
   */
  public static void println(String s) { print(s);  println(); }

  /**
   *
   */
  public static void println() { print("\n"); }

  /**
   *
   * @return
   * @throws IOException
   */
  public static String readLine() throws IOException {
    return self.cp.model.readLine();
  }

  /**
   *
   */
  public static void clear() {
    self.cp.model.content.clear();
    self.cp.model.maxLine = "";
    self.cp.revalidate();
  }

}


/**
 *
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: FEL CVUT</p>
 * @author Jiri Danecek
 * @version 1.0
 */
class ApplicationPanel extends Box {

  Thread mainThread;
  Method main;
  Class mainClass;
  Console console;
  private static boolean cancelFl = false;

  JButton startButton;
  JButton stopButton;

  javax.swing.Timer timer;

      class MyButton extends JButton {

        Dimension dim;

        MyButton(MyAction myAction) {
          super(myAction);
          setToolTipText(myAction.getTip());
          if (myAction.icon != null)
            dim = new Dimension(myAction.icon.getIconWidth(), myAction.icon.getIconHeight());
        }

        public Dimension getPreferredSize() {
          if (dim != null) return dim;
          return super.getPreferredSize();
        }

        public Dimension getMaximumSize() { return getPreferredSize();}
      }

      class StartAction extends MyAction {

        StartAction() {
          super(new String[]{"", "Start"}, new ImageIcon(ApplicationPanel.class.getResource("resources/run.gif")));
        }

        public void actionPerformed(ActionEvent e) {
           startButton.setEnabled(false);
           stopButton.setEnabled(true);
           mainThread = new Thread(
              new Runnable() {
                public void run() {
                   try { main.invoke(null, new Object[] {null});}
                   catch (InvocationTargetException e)  {
                     System.out.println("Program was terminated by the cause: "  + e.getCause());}
                   catch (RuntimeException re)       { System.out.println(re); }
                   catch (IllegalAccessException e)  { e.printStackTrace(); }
                   end();
                }
              }
           );
           mainThread.setPriority(Thread.MIN_PRIORITY);
           mainThread.start();
        }
      }

      class StopAction extends MyAction {

        StopAction() {
          super(new String[]{"", "Stop "}, new ImageIcon(ApplicationPanel.class.getResource("resources/reset.gif")));
        }

        public void actionPerformed(ActionEvent e) {
          if (mainThread != null) {
            mainThread.interrupt();
            timer.start();
          }
        }
      }

      class ClearAction extends MyAction {

        ClearAction() {
          super(new String[]{ "", "Vyma"},
          new ImageIcon(ApplicationPanel.class.getResource("resources/clear.gif")));
        }


        public void actionPerformed(ActionEvent e) {
          Console.clear();
        }
      }

  private void end() {
    if (timer != null && timer.isRunning())  timer.stop();
    Display.free();
    mainThread = null;
    startButton.setEnabled(true);
    stopButton.setEnabled(false);
  }

  public static boolean isApplicationCanceled() { return cancelFl; }

  ApplicationPanel(String mainClassName) throws ClassNotFoundException, NoSuchMethodException {

    super(BoxLayout.Y_AXIS);

//    setBorder(BorderFactory.createBevelBorder(BevelBorder.LOWERED));
//    setBorder(BorderFactory.createTitledBorder("ApplicationPanel"));

    timer = new javax.swing.Timer(3000,
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          if (mainThread != null) mainThread.stop();
          end();
        }
      }
    );
    timer.setRepeats(false);

    mainClass = Class.forName(mainClassName);
    main = mainClass.getMethod("main", new Class[] { String[].class} );
    add(console = Console.createConsole(mainClassName));

    Box buttonBox = new Box(BoxLayout.X_AXIS);
    int w = 3;
    stopButton = new MyButton(new StopAction());
    startButton = new MyButton(new StartAction());
    buttonBox.add(Box.createHorizontalStrut(w));
    buttonBox.setMaximumSize(new Dimension(10000, stopButton.getHeight()));
    buttonBox.add(startButton);
    buttonBox.add(Box.createHorizontalStrut(w));
    buttonBox.add(stopButton);
    buttonBox.add(Box.createHorizontalGlue());
    buttonBox.add(new MyButton(new ClearAction()));
    buttonBox.add(Box.createHorizontalStrut(w));
    stopButton.setEnabled(false);
    add(buttonBox);
  }

}
