package sugar;
import java.io.*;

/** This class enables simple data entry from console ( System.in )
 *  for Strings, double and integer numbers.
 * @author J.M.Bloudil & Jiri Danecek
 * @version 2.0
 */

public class Sys {

  static BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
  static private String  last=null;      // last element precessed
  static private boolean error;          // error condition
  static private boolean end;            // end condition
  static private String buf = null;      // buffer
  static private int curInd, lastInd;    // gliding indices
  static private String byeMsg = "#";    // ending input message
  static private String inData;

  /**
   * Read one word ie. continuous sequence of letters
   * @return word or zero length string on end.
   */
  static public String readWord() {
    error=false;
    if (end) return "";

    do {
      if (buf==null || buf.equals("")) {
        do {
          readBuf(true);           // fill buffer and detect first and last letter
          if (end) return "";
          while (curInd<=lastInd && !Character.isLetter(buf.charAt(curInd)))  curInd++;
          while (curInd<=lastInd && !Character.isLetter(buf.charAt(lastInd))) lastInd--;
        } while (curInd>lastInd);
      }
      while (curInd<=lastInd && !Character.isLetter(buf.charAt(curInd))) curInd++;
    } while (curInd>lastInd);
    int k = curInd+1;
    while (k<=lastInd && Character.isLetter(buf.charAt(k))) k++;
    String res = buf.substring(curInd, k);
    curInd = k;
    if (curInd>lastInd) buf = "";
    return last=res;
  }

  /**
   * Read one character
   * @return one character or blank (on end) or '\n' (on empty buffer)
   */
  static public char readChar() {
    if (buf==null) readBuf(false);                 // fill buffer
    last=buf;                                 // last read element
    if (end) return ' ';
    if (buf.equals("") || curInd>lastInd) { buf = null; return '\n'; }
    return buf.charAt(curInd++);
  }


  /** The method reads one input line. If line equals bye string
   *  end condition is raised. Prompt is eventually printed.
   *  @return string (without enter character).
   */
  public static String readLine() {
    error=false;
    String s=null;
    if (end) error=true;
    try {
      s=(Console.isInitialized())? Console.readLine() : br.readLine();
      if ( s==null || s.equals(byeMsg) ) end=true;
    } catch (IOException e) {System.out.println("IO Exc"); error=true;}
    return last=s;
  }

  /** The method reads one number in one input line. If line equals bye string
   *  end condition is raised. Prompt is eventually printed.
   *  If input is bye message 0 is returned and end condition is raised.
   *  If input is not of form of number error condition is raised and NaN is
   *  returned. Entered number can be preceded or followed by blanks and tabs.
   *  @return double.
   */
  public static double readDouble() {
     double d = Double.NaN;
     last = readToken();
     if (end) return d;
     return getDouble(last);
  }

  /** The method reads one number in one input line. If line equals bye string
   *  end condition is raised. Prompt is eventually printed.
   *  If input is bye message, 0 is returned and end condition is raised.
   *  If input does not represent integer, ie. does not to int, its fraction part
   *  is not zero, error condition is raised.
   *  Entered number can be preceded or followed by blanks and tabs.
   *  @return integer part of entered number or 0 on bye.
   */
  public static int readInt() {
     last = readToken();
     if (end) return Integer.MIN_VALUE;
     return getInt(last);
  }

  /**  Print overloaded methods
   *
   */
  public static void p(boolean b){p("" + b);}
  public static void p(char c)   {p("" + c);}
  public static void p(char[] s) {p("" + s);}
  public static void p(double d) {p("" + d);}
  public static void p(float f)  {p("" + f);}
  public static void p(int i)    {p("" + i);}
  public static void p(long l)   {p("" + l);}
  public static void p(Object o) {p("" + o);}
  public static void p(double d, int fract) { p(Num.form(d,fract)); }
  public static void p(String s) {
    if (Console.isInitialized()) Console.print(s);
    else System.out.print(s);
  }

  /**  Print next line overloaded methods
   *
   */
  public static void pln()         {pln("");}
  public static void pln(boolean b){pln("" + b);}
  public static void pln(char c)   {pln("" + c);}
  public static void pln(char[] s) {pln("" + s);}
  public static void pln(double d) {pln("" + d);}
  public static void pln(float f)  {pln("" + f);}
  public static void pln(int i)    {pln("" + i);}
  public static void pln(long l)   {pln("" + l);}
  public static void pln(Object o) {pln("" + o);}
  public static void pln(double d, int fract) { pln(Num.form(d, fract)); }
  public static void pln(String s) {
    if (Console.isInitialized()) Console.println(s);
    else System.out.println(s);
  }

  /**
   * Sleeps program for msec milisecond
   */
  public static void sleep(long msec) {
    try { Thread.sleep(msec); }
    catch (InterruptedException ex) {
      throw new RuntimeException("Interrupted in sleep: " + ex);
    }
  }

  /*  Fills input buffer by given string and appends one blank and bye message;
   *  @param input string
   */
  static public void setInputData(String s) { inData=s; }

  /* Get status information.
   * @return string information.
   */
  static public String getState() {
    return "Sys[last="+last+", error="+error+", end="+end+"]";
  }

  /**
   * Sets end of input message.
   * @param msg - the ending message.
   */
  static public void setByeMsg( String msg ) {
    byeMsg = msg;
  }

  /**
   * Gets end of input message.
   * @return the message
   */
  static public String getByeMsg() {
    return byeMsg;
  }

  /**
   * Get last processed element.
   * @return last element
   */
  static public String getLast() {
    return last;
  }

  /**
   *   Read one token splited by blank(s).
   * @return token or null.
   */
  static private String readToken() {
    error=false;
    do {
      if (end) return null;
      //System.out.println("to");
      if (buf==null || buf.equals(""))  readBuf(true);  // fill trimed buffer
      while (curInd<=lastInd && buf.charAt(curInd)==' ') curInd++;  // skip
    } while (curInd>lastInd);                           //
    int k = curInd+1;
    while (k<=lastInd && buf.charAt(k)!=' ') k++;
    String res = buf.substring(curInd, k);
    curInd = k;
    if (curInd>lastInd) last=buf = "";
    if (res.equals(byeMsg)) { end=true; return null; }
    return last=res;
  }

  /**
   * Indicates whether error condition is raised and reset it to false.
   * @return true means error.
   */
  public static boolean  getError() { return error; }

  /** The method gives end status.
   *  @return true means end condition.
   */
  public static boolean  getEnd() {return end;}

  /** The method sets end status.
   *  @param false or true.
   */
  public static void     setEnd(boolean e) {end=e;}

  /** The method reads one input line. If line equals bye string
   *  end condition is raised. Prompt is eventually printed.
   *  @param  query message to user to stop input.
   *  @return true if user's reply equals to bye message, false - otherwise.
   */

  public static boolean askStop(String query) {
     System.out.print(query);
     readLine();
     return getEnd();
  }

  /** The method reads one number from string.
   *  If input does not represent integer, ie. does not to int, its fraction part
   *  is not zero, error condition is raised.
   *  Entered number can be preceded or followed by blanks and tabs.
   *  @return integer part of entered number or 0 on bye.
   */
  private static int getInt(String s) {
     last=s;
     error=false;
     end=false;
     double d=getDouble(s);
     int i=new Double(d).intValue();
     if (d!=i) {error=true; return Integer.MIN_VALUE; }
     return i;
  }

 /** The method reads one number from string.
   *  If input is not of form of number error condition is raised and NaN is
   *  returned. Entered number can be preceded or followed by blanks and tabs.
   *  @return double.
   */
  public static double getDouble(String s) {
     last=s;
     double d=Double.NaN;
     end=false;
     error=false;
     if (s==null || s.length()==0 ) {error=true; return d;}
     try {d=Double.parseDouble(s);}
     catch(NumberFormatException e) { error=true;}
     return d;
  }

//-----------------------------------------------
//---------------- private part -----------------
//-----------------------------------------------

  /**
   * Reads line from buffered reader or from Console
   */
  static private void readBuf(boolean trim) {
    error=false;
    try {
      if (inData!=null) { last=buf=inData; inData=null; }
      else {
        if (Console.isInitialized())
          last=buf=Console.readLine();
        else
          last=buf=br.readLine();
      }
    } catch (IOException e) { buf = ""; error=true; end = true; } // ERROR

    if (buf.equals(byeMsg)) { buf = ""; end = true; }             // END
    else {
      if (trim) buf=buf.trim();
      curInd = 0;
      lastInd = buf.length()-1;
    }                                                             // OK
  }

  // TESTER ********************************************************************

  public static void main(String[] args) {

    String s="";

    System.out.println(Sys.getDouble("  12.1 "));

    System.out.println(Sys.getByeMsg());

 //   Sys.setInputData("11 aa 22 2e2");

    int k=Sys.getInt("222");

    System.out.println(Sys.getState());

    if (Sys.getEnd()) return;
    System.out.println(k);

    System.out.println("********************");

    while (true) {
       double  d=Sys.readDouble();
       if (Sys.getEnd()) break;
       if (Sys.getError()) Sys.pln("CHYBA ...  >"+Sys.getState()+"<   ");
       else Sys.pln(d);
    }
  }
/*
  public static void main(String[] args) {
    char c = readChar();
    while (!getEnd()) {
      p(c); c = readChar();
    }
    pln("konec");
  }
*/

/*
  public static void main(String[] args) {
    String s = readToken();
    while (!getEnd()) {
      pln(s); s = readToken();
    }
    pln("konec");
  }
*/
  /*
  public static void main(String[] args) {
    String x = readWord();
    while (!getEnd()) {
      pln(x); x = readWord();
    }
    pln("konec");
  }
  */


}